/**
 * Created by AlexanderC on 6/15/15.
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.DB = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _deepKernel = require('deep-kernel');

var _deepKernel2 = _interopRequireDefault(_deepKernel);

var _deepCore = require('deep-core');

var _deepCore2 = _interopRequireDefault(_deepCore);

var _vogels = require('vogels');

var _vogels2 = _interopRequireDefault(_vogels);

var _ExtendModel = require('./Vogels/ExtendModel');

var _ModelNotFoundException = require('./Exception/ModelNotFoundException');

var _deepValidation = require('deep-validation');

var _deepValidation2 = _interopRequireDefault(_deepValidation);

var _util = require('util');

var _util2 = _interopRequireDefault(_util);

var _FailedToCreateTableException = require('./Exception/FailedToCreateTableException');

var _FailedToCreateTablesException = require('./Exception/FailedToCreateTablesException');

var _AbstractDriver = require('./Local/Driver/AbstractDriver');

var _AutoScaleDynamoDB = require('./DynamoDB/AutoScaleDynamoDB');

var _EventualConsistency = require('./DynamoDB/EventualConsistency');

var _https = require('https');

var _https2 = _interopRequireDefault(_https);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Vogels wrapper
 */
let DB = exports.DB = function (_Kernel$ContainerAwar) {
  _inherits(DB, _Kernel$ContainerAwar);

  /**
   * @param {Array} models
   * @param {Object} tablesNames
   * @param {Boolean} forcePartitionField
   * @param {String[]} nonPartitionedModels
   */
  function DB() {
    let models = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    let tablesNames = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    let forcePartitionField = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
    let nonPartitionedModels = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : [];

    _classCallCheck(this, DB);

    var _this = _possibleConstructorReturn(this, (DB.__proto__ || Object.getPrototypeOf(DB)).call(this));

    _this._tablesNames = tablesNames;
    _this._forcePartitionField = forcePartitionField;
    _this._dynamoDBPartitionKey = null;

    _this._validation = new _deepValidation2.default(models, forcePartitionField, nonPartitionedModels);
    _this._rawModels = _this._rawModelsVector(models);
    _this._models = {};
    _this._workingCreds = null;
    return _this;
  }

  /**
   * @returns {Validation}
   */


  _createClass(DB, [{
    key: 'has',


    /**
     * @param {String} modelName
     * @returns {Boolean}
     */
    value: function has(modelName) {
      return this._rawModels.indexOf(modelName) !== -1;
    }

    /**
     * @param {String} modelName
     * @returns {Vogels}
     */

  }, {
    key: 'get',
    value: function get(modelName) {
      if (!this.has(modelName)) {
        modelName = this._lispCase(modelName);

        if (!this.has(modelName)) {
          throw new _ModelNotFoundException.ModelNotFoundException(modelName);
        }
      }

      if (!this._models.hasOwnProperty(modelName)) {
        this._models[modelName] = this._rawModelToVogels(modelName);
      }

      this._ensureModelPartitioned(modelName);
      this._ensureModelCredentials(modelName);

      let model = this._models[modelName];

      // inject logService into extended model to log RUM events
      if (this.kernel && this.kernel.isRumEnabled && !model.hasOwnProperty('logService')) {
        model.logService = this.kernel.get('log');
      }

      return model;
    }

    /**
     * @param {String} modelName
     * @param {Function} callback
     * @param {Object} options
     * @returns {DB}
     */

  }, {
    key: 'assureTable',
    value: function assureTable(modelName, callback) {
      let options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

      if (!this.has(modelName)) {
        throw new _ModelNotFoundException.ModelNotFoundException(modelName);
      }

      this.get(modelName); // ensure DynamoDB model initialized
      options = _util2.default._extend(DB.DEFAULT_TABLE_OPTIONS, options);
      options[modelName] = options;

      _vogels2.default.createTables(options, error => {
        if (error) {
          throw new _FailedToCreateTableException.FailedToCreateTableException(modelName);
        }

        callback();
      });

      return this;
    }

    /**
     * @param {Function} callback
     * @param {Object} options
     * @returns {DB}
     */

  }, {
    key: 'assureTables',
    value: function assureTables(callback) {
      let options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      let allModelsOptions = {};
      let allModelNames = [];

      for (let modelName in this.models) {
        if (!this.models.hasOwnProperty(modelName)) {
          continue;
        }

        allModelsOptions[modelName] = _util2.default._extend(DB.DEFAULT_TABLE_OPTIONS, options.hasOwnProperty(modelName) ? options[modelName] : {});
        allModelNames.push(modelName);
      }

      _vogels2.default.createTables(allModelsOptions, error => {
        if (error) {
          throw new _FailedToCreateTablesException.FailedToCreateTablesException(allModelNames, error);
        }

        callback();
      }, this._localBackend);

      return this;
    }

    /**
     * Booting a certain service
     *
     * @param {Kernel} kernel
     * @param {Function} callback
     */

  }, {
    key: 'boot',
    value: function boot(kernel, callback) {
      this._validation.kernel = kernel;

      this._validation.boot(kernel, () => {
        this._tablesNames = kernel.config.tablesNames;
        this._rawModels = this._rawModelsVector(kernel.config.models);

        if (this._localBackend) {
          this._enableLocalDB(() => {
            if (!_vogels2.default.documentClient().hasOwnProperty(_EventualConsistency.EventualConsistency.DEEP_DB_DECORATOR_FLAG)) {
              this._initEventualConsistency(kernel);
            }

            callback();
          });
        } else {
          this._fixNodeHttpsIssue();

          // it's important to be loaded before any other decorator
          if (!_vogels2.default.documentClient().hasOwnProperty(_EventualConsistency.EventualConsistency.DEEP_DB_DECORATOR_FLAG)) {
            this._initEventualConsistency(kernel);
          }

          if (!_vogels2.default.documentClient().hasOwnProperty(_AutoScaleDynamoDB.AutoScaleDynamoDB.DEEP_DB_DECORATOR_FLAG)) {
            this._initVogelsAutoscale(kernel);
          }

          callback();
        }
      });
    }

    /**
     * @inheritDoc
     */

  }, {
    key: 'cleanup',
    value: function cleanup() {
      let vogelsDynamoDriver = this.vogelsDynamoDriver;

      if (vogelsDynamoDriver.config.httpOptions.agent) {
        vogelsDynamoDriver.config.httpOptions.agent.destroy();
      }

      this._fixNodeHttpsIssue();

      // restore env credentials when finishing lambda execution
      // to avoid security token expired exception on successive calls
      if (_deepCore2.default.AWS.ENV_CREDENTIALS) {
        this.overwriteCredentials(_deepCore2.default.AWS.ENV_CREDENTIALS);
      }
    }

    /**
     * NetworkingError: write EPROTO
     *
     * @see https://github.com/aws/aws-sdk-js/issues/862
     * @private
     */

  }, {
    key: '_fixNodeHttpsIssue',
    value: function _fixNodeHttpsIssue() {
      _vogels2.default.AWS.config.maxRetries = 3;

      this._setVogelsDriver(new _vogels2.default.AWS.DynamoDB({
        maxRetries: 8,
        retryDelayOptions: {
          base: 100
        },
        httpOptions: {
          agent: new _https2.default.Agent({
            rejectUnauthorized: true,
            keepAlive: true,
            secureProtocol: 'TLSv1_method',
            ciphers: 'ALL'
          })
        }
      }));
    }

    /**
     * @param {Kernel} kernel
     *
     * @private
     */

  }, {
    key: '_initEventualConsistency',
    value: function _initEventualConsistency(kernel) {
      _vogels2.default.documentClient(new _EventualConsistency.EventualConsistency(_vogels2.default.dynamoDriver(), _vogels2.default.documentClient(), kernel, this._models).localMode(this._localBackend).extend());
    }

    /**
     * @param {Kernel} kernel
     *
     * @private
     */

  }, {
    key: '_initVogelsAutoscale',
    value: function _initVogelsAutoscale(kernel) {
      _vogels2.default.documentClient(new _AutoScaleDynamoDB.AutoScaleDynamoDB(_vogels2.default.dynamoDriver(), _vogels2.default.documentClient(), kernel).extend());
    }

    /**
     * @param {Object} driver
     * @returns {DB}
     * @private
     */

  }, {
    key: '_setVogelsDriver',
    value: function _setVogelsDriver(driver) {
      _vogels2.default.dynamoDriver(driver);

      return this;
    }

    /**
     * @returns {AWS.DynamoDB}
     */

  }, {
    key: 'overwriteCredentials',


    /**
     * @param {AWS.Credentials} credentials
     * @returns {DB}
     */
    value: function overwriteCredentials(credentials) {
      this._workingCreds = credentials;

      let dynamoDriver = _vogels2.default.dynamoDriver();
      let docClient = _vogels2.default.documentClient();

      docClient.service.config.credentials = credentials;
      dynamoDriver.config.credentials = credentials;

      // update docClient credentials for each model
      for (let modelName in this._models) {
        if (!this._models.hasOwnProperty(modelName)) {
          continue;
        }

        this._ensureModelCredentials(modelName);
      }

      return this;
    }

    /**
     * @param {Function} callback
     * @param {String} driver
     * @param {Number} tts
     * @returns {AbstractDriver}
     */

  }, {
    key: '_enableLocalDB',


    /**
     * @param {Function} callback
     * @private
     */
    value: function _enableLocalDB(callback) {
      this._setVogelsDriver(this._localDynamoDb);

      this.assureTables(callback);
    }

    /**
     * @returns {Object}
     */

  }, {
    key: '_rawModelsVector',


    /**
     * @param {Array} rawModels
     * @returns {String[]}
     */
    value: function _rawModelsVector(rawModels) {
      let modelsVector = [];

      for (let modelKey in rawModels) {
        if (!rawModels.hasOwnProperty(modelKey)) {
          continue;
        }

        let backendModels = rawModels[modelKey];

        for (let modelName in backendModels) {
          if (!backendModels.hasOwnProperty(modelName)) {
            continue;
          }

          modelsVector.push(modelName);
        }
      }

      return modelsVector;
    }

    /**
     * @param {String} modelName
     *
     * @returns {*}
     *
     * @private
     */

  }, {
    key: '_rawModelToVogels',
    value: function _rawModelToVogels(modelName) {
      return new _ExtendModel.ExtendModel(_vogels2.default.define(modelName, this._wrapModelSchema(modelName))).inject();
    }

    /**
     * @param {Array} rawModels
     * @returns {Object}
     */

  }, {
    key: '_rawModelsToVogels',
    value: function _rawModelsToVogels(rawModels) {
      let models = {};

      this._rawModelsVector(rawModels).map(modelName => {
        models[modelName] = this._rawModelToVogels(modelName);
      });

      return models;
    }

    /**
     * @param {String} modelName
     *
     * @private
     */

  }, {
    key: '_ensureModelCredentials',
    value: function _ensureModelCredentials(modelName) {
      this._workingCreds = this._workingCreds || _deepCore2.default.AWS.ENV_CREDENTIALS;

      if (this._workingCreds && this._models.hasOwnProperty(modelName)) {
        this._models[modelName].docClient.service.config.credentials = this._workingCreds;
      }
    }

    /**
     * @param {String} modelName
     * 
     * @private
     */

  }, {
    key: '_ensureModelPartitioned',
    value: function _ensureModelPartitioned(modelName) {
      if (this._models.hasOwnProperty(modelName) && this.validation.isPartitionedModel(modelName)) {

        this._models[modelName].setPartition(this._dynamoDBPartitionKey || _ExtendModel.ExtendModel.ANONYMOUS_PARTITION);
      }
    }

    /**
     * @param {String} partitionKey
     * @returns {DB}
     */

  }, {
    key: 'setDynamoDBPartitionKey',
    value: function setDynamoDBPartitionKey(partitionKey) {
      this._dynamoDBPartitionKey = partitionKey;

      Object.keys(this._models).map(modelName => {
        this._ensureModelPartitioned(modelName);
      });

      return this;
    }

    /**
     * @param {String} name
     * @returns {Object}
     * @private
     */

  }, {
    key: '_wrapModelSchema',
    value: function _wrapModelSchema(name) {
      let schema = {
        timestamps: true,
        tableName: this._tablesNames[name],
        schema: this._validation.getSchema(name)
      };

      if (this._usePartitionField && this.validation.isPartitionedModel(name)) {
        schema.hashKey = _ExtendModel.ExtendModel.PARTITION_FIELD;
        schema.rangeKey = DB.DEFAULT_KEY_FIELD;
      } else {
        schema.hashKey = DB.DEFAULT_KEY_FIELD;
      }

      return schema;
    }

    /**
     * @returns {Boolean}
     * @private
     */

  }, {
    key: '_lispCase',


    /**
     * @param {String} str
     * @returns {String}
     * @private
     */
    value: function _lispCase(str) {
      return str.replace(/([a-z])([A-Z])/g, '$1-$2').split(/[^a-z0-9\-]+/i).join('-').toLowerCase();
    }

    /**
     * @returns {*}
     * @private
     */

  }, {
    key: 'validation',
    get: function get() {
      return this._validation;
    }

    /**
     * @returns {String[]}
     */

  }, {
    key: 'rawModels',
    get: function get() {
      return this._rawModels;
    }

    /**
     * @returns {Vogels[]}
     */

  }, {
    key: 'models',
    get: function get() {
      if (this._rawModels.length <= Object.keys(this._models).length) {
        return this._models;
      }

      let result = {};

      this._rawModels.map(modelName => {
        result[modelName] = this.get(modelName);
      });

      return result;
    }
  }, {
    key: 'vogelsDynamoDriver',
    get: function get() {
      return _vogels2.default.dynamoDriver();
    }
  }, {
    key: '_localDynamoDb',


    /**
     * @returns {AWS.DynamoDB|VogelsMock.AWS.DynamoDB|*}
     * @private
     */
    get: function get() {
      return new _vogels2.default.AWS.DynamoDB({
        endpoint: new _vogels2.default.AWS.Endpoint(`http://localhost:${DB.LOCAL_DB_PORT}`),
        accessKeyId: 'fake',
        secretAccessKey: 'fake',
        region: 'us-east-1'
      });
    }
  }, {
    key: '_usePartitionField',
    get: function get() {
      return this._forcePartitionField || this.kernel && this.kernel.accountMicroservice;
    }
  }, {
    key: '_security',
    get: function get() {
      return this.container.get('security');
    }

    /**
     * @returns {String}
     */

  }], [{
    key: 'startLocalDynamoDBServer',
    value: function startLocalDynamoDBServer(callback) {
      let driver = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'LocalDynamo';
      let tts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : _AbstractDriver.AbstractDriver.DEFAULT_TTS;

      let LocalDBServer = require('./Local/DBServer').DBServer;

      let server = LocalDBServer.create(driver);

      server.start(callback, tts);

      return server;
    }
  }, {
    key: 'DEFAULT_TABLE_OPTIONS',
    get: function get() {
      return {
        readCapacity: 1,
        writeCapacity: 1
      };
    }
  }, {
    key: 'DEFAULT_KEY_FIELD',
    get: function get() {
      return 'Id';
    }

    /**
     * @returns {Number}
     */

  }, {
    key: 'LOCAL_DB_PORT',
    get: function get() {
      return _AbstractDriver.AbstractDriver.DEFAULT_PORT;
    }
  }]);

  return DB;
}(_deepKernel2.default.ContainerAware);