/**
 * Created by AlexanderC on 6/08/16.
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.EventualConsistency = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _util = require('util');

var _awsSdk = require('aws-sdk');

var _awsSdk2 = _interopRequireDefault(_awsSdk);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

let EventualConsistency = exports.EventualConsistency = function () {
  /**
   * @param {AWS.DynamoDB|AWS.DynamoDB.DocumentClient|*} dynamoDb
   * @param {AWS.DynamoDB|AWS.DynamoDB.DocumentClient|*} dynamoDbDocumentClient
   * @param {Kernel} deepKernel
   * @param {Object} modelsToExtend
   */
  function EventualConsistency(dynamoDb, dynamoDbDocumentClient, deepKernel, modelsToExtend) {
    _classCallCheck(this, EventualConsistency);

    this._dynamoDb = dynamoDb;
    this._dynamoDbDocumentClient = dynamoDbDocumentClient;
    this._kernel = deepKernel;
    this._modelsToExtend = modelsToExtend;
    this._localModel = false;
    this._sqsInstances = {};
  }

  /**
   * @returns {Object}
   */


  _createClass(EventualConsistency, [{
    key: '_sqs',


    /**
     * @param {String} modelName
     * @returns {AWS.SQS|*}
     */
    value: function _sqs(modelName) {
      if (!this._sqsInstances.hasOwnProperty(modelName)) {
        let queuesMapping = this._queuesModelMapping;

        if (!queuesMapping.hasOwnProperty(modelName)) {
          return null;
        }

        this._sqsInstances[modelName] = new _awsSdk2.default.SQS({
          region: EventualConsistency.getRegionFromSqsQueueUrl(queuesMapping[modelName].url)
        });
      }

      return this._sqsInstances[modelName];
    }

    /**
     * @param {String} modelName
     * @param {String} method
     * @param {Object} payload
     * @param {Function} callback
     * @returns {*}
     * @private
     */

  }, {
    key: '_sendPayload',
    value: function _sendPayload(modelName, method, payload, callback) {
      let queuesMapping = this._queuesModelMapping;

      if (!queuesMapping.hasOwnProperty(modelName)) {
        return callback(null);
      }

      let params = {
        MessageBody: JSON.stringify({ method: method, payload: payload }),
        QueueUrl: queuesMapping[modelName].url
      };

      this._sqs(modelName).sendMessage(params, error => callback(error));
    }

    /**
     * @param {String} queueUrl
     * @returns {String}
     */

  }, {
    key: 'localMode',


    /**
     * @param {Boolean} state
     * @returns {EventualConsistency}
     */
    value: function localMode(state) {
      this._localModel = !!state;

      return this;
    }

    /**
     * @returns {AWS.DynamoDB|AWS.DynamoDB.DocumentClient|*}
     */

  }, {
    key: 'extend',
    value: function extend() {

      // avois doing it for older deploys
      if (this._kernel.config.dbOffloadQueues) {
        EventualConsistency.PROXIED_METHODS.forEach(method => {
          let originalMethod = `__deep_db_EC_${method}__`;

          this._dynamoDbDocumentClient[originalMethod] = this._dynamoDbDocumentClient[method];
          this._dynamoDbDocumentClient[method] = this._ecExecCb(method, originalMethod);
        });

        this._extendModels();
        this._dynamoDbDocumentClient[EventualConsistency.DEEP_DB_DECORATOR_FLAG] = true;
      }

      return this._dynamoDbDocumentClient;
    }

    /**
     * @param {String} modelName
     * @param {*} model
     *
     * @returns {EventualConsistency|*}
     */

  }, {
    key: 'extendModel',
    value: function extendModel(modelName, model) {
      this._modelsToExtend[modelName] = model;

      this._extendModels();

      return this;
    }

    /**
     * @private
     */

  }, {
    key: '_extendModels',
    value: function _extendModels() {
      let queuesMapping = this._queuesModelMapping;
      let tablesNames = this._kernel.config.tablesNames;

      for (let modelName in this._modelsToExtend) {
        if (!this._modelsToExtend.hasOwnProperty(modelName) || this._modelsToExtend[modelName].hasOwnProperty(EventualConsistency.DEEP_DB_EC_MODEL_NAME_PROPERTY) || !queuesMapping.hasOwnProperty(modelName)) {
          // weird case that should be covered
          continue;
        }

        let model = this._modelsToExtend[modelName];

        model[EventualConsistency.DEEP_DB_EC_MODEL_NAME_PROPERTY] = tablesNames[modelName];
        model[EventualConsistency.DEEP_DB_EC_STATE_PROPERTY] = false;
        model[EventualConsistency.DEEP_DB_EC_STATE_PROPERTY_SETTER] = state => {
          model[EventualConsistency.DEEP_DB_EC_STATE_PROPERTY] = !!state;

          return model;
        };
      }
    }

    /**
     * @param {String} method
     * @param {String} originalMethod
     * @returns {Function}
     */

  }, {
    key: '_ecExecCb',
    value: function _ecExecCb(method, originalMethod) {
      return (payload, originalCb) => {
        let tableName = EventualConsistency._getTableNameFromPayload(method, payload);

        this._originalExec(method, originalMethod, payload, originalCb, this._isEnabledForTable(tableName) ? () => {
          if (this._localModel) {
            process.nextTick(() => {

              //todo: remove this log?
              console.log(`DynamoDB EC->${method}(`, JSON.stringify(payload), ')');

              this._dynamoDbDocumentClient[originalMethod]((0, _util._extend)({}, payload), () => {});
            });

            return originalCb(null, { Attributes: {} }); // assume that no errors occured while writing to SQS
          }

          this._sendPayload(this._modelName(tableName), method, payload, error => originalCb(error, { Attributes: {} }));
        } : null);
      };
    }

    /**
     * @param {String} method
     * @param {String} originalMethod
     * @param {Object} payload
     * @param {Function} originalCb
     * @param {Function} onOpFailedCb
     * @private
     */

  }, {
    key: '_originalExec',
    value: function _originalExec(method, originalMethod, payload, originalCb) {
      let onOpFailedCb = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : null;

      this._dynamoDbDocumentClient[originalMethod](payload, (error, data) => {
        if (error && onOpFailedCb && EventualConsistency.ERRORS_TO_SKIP.indexOf(error.code) === -1) {

          return onOpFailedCb();
        }

        originalCb(error, data);
      });
    }

    /**
     * @returns {String[]}
     */

  }, {
    key: '_modelName',


    /**
     * @param {String} tableName
     * @returns {String}
     */
    value: function _modelName(tableName) {
      for (let modelName in this._modelsToExtend) {
        if (!this._modelsToExtend.hasOwnProperty(modelName)) {
          continue;
        }

        let model = this._modelsToExtend[modelName];

        if (model[EventualConsistency.DEEP_DB_EC_MODEL_NAME_PROPERTY] === tableName) {
          return modelName;
        }
      }

      return null;
    }

    /**
     * @param {String} tableName
     * @returns {Boolean}
     */

  }, {
    key: '_isEnabledForTable',
    value: function _isEnabledForTable(tableName) {
      for (let modelName in this._modelsToExtend) {
        if (!this._modelsToExtend.hasOwnProperty(modelName)) {
          continue;
        }

        let model = this._modelsToExtend[modelName];

        if (model[EventualConsistency.DEEP_DB_EC_MODEL_NAME_PROPERTY] === tableName) {
          if (model[EventualConsistency.DEEP_DB_EC_STATE_PROPERTY]) {
            return true;
          }

          break;
        }
      }

      return false;
    }

    /**
     * @param {String} str
     * @returns {String}
     */

  }, {
    key: '_queuesModelMapping',
    get: function get() {
      return this._kernel.config.dbOffloadQueues || {};
    }
  }, {
    key: 'modelsToExtend',


    /**
     * @returns {Object}
     */
    get: function get() {
      return this._modelsToExtend;
    }

    /**
     * @returns {AWS.DynamoDB|AWS.DynamoDB.DocumentClient|*}
     */

  }, {
    key: 'dynamoDbDocumentClient',
    get: function get() {
      return this._dynamoDbDocumentClient;
    }

    /**
     * @returns {AWS.DynamoDB|AWS.DynamoDB.DocumentClient|*}
     */

  }, {
    key: 'dynamoDb',
    get: function get() {
      return this._dynamoDb;
    }

    /**
     * @returns {String[]}
     */

  }], [{
    key: 'getRegionFromSqsQueueUrl',
    value: function getRegionFromSqsQueueUrl(queueUrl) {
      let regionParts = queueUrl.match(/\.([^\.]+)\.amazonaws\.com\/.*/i);

      if (!regionParts || regionParts.length === 0) {
        throw new Error(queueUrl, 'Unable to extract AWS region.');
      }

      return regionParts[1];
    }
  }, {
    key: '_capitalizeFirst',
    value: function _capitalizeFirst(str) {
      return str.charAt(0).toUpperCase() + str.slice(1);
    }

    /**
     * @param {String} method
     * @param {Object} payload
     * @returns {String}
     * @private
     */

  }, {
    key: '_getTableNameFromPayload',
    value: function _getTableNameFromPayload(method, payload) {
      let tableName = null;

      switch (method) {
        case 'put':
        case 'update':
          tableName = payload.TableName;
          break;
        default:
          throw new Error(`Unknown DynamoDB eventual consistency method '${method}'`);
      }

      return tableName;
    }
  }, {
    key: 'ERRORS_TO_SKIP',
    get: function get() {
      return ['ResourceNotFoundException', 'ItemCollectionSizeLimitExceededException', 'ConditionalCheckFailedException'];
    }
  }, {
    key: 'PROXIED_METHODS',
    get: function get() {
      return ['put', 'update'];
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'DEEP_DB_EC_STATE_PROPERTY',
    get: function get() {
      return '_eventualConsistency';
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'DEEP_DB_EC_MODEL_NAME_PROPERTY',
    get: function get() {
      return '__deep_ec_short_model_name__';
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'DEEP_DB_EC_STATE_PROPERTY_SETTER',
    get: function get() {
      return 'eventualConsistency';
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'DEEP_DB_DECORATOR_FLAG',
    get: function get() {
      return '__deep_db_ec_decorator__';
    }
  }]);

  return EventualConsistency;
}();