/**
 * Created by AlexanderC on 9/23/15.
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.AbstractDriver = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _deepCore = require('deep-core');

var _deepCore2 = _interopRequireDefault(_deepCore);

var _ServerAlreadyRunningException = require('./Exception/ServerAlreadyRunningException');

var _ServerTtsExceededException = require('./Exception/ServerTtsExceededException');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

let AbstractDriver = exports.AbstractDriver = function (_Core$OOP$Interface) {
  _inherits(AbstractDriver, _Core$OOP$Interface);

  /**
   * @param {Number} port
   */
  function AbstractDriver() {
    let port = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : AbstractDriver.DEFAULT_PORT;

    _classCallCheck(this, AbstractDriver);

    var _this = _possibleConstructorReturn(this, (AbstractDriver.__proto__ || Object.getPrototypeOf(AbstractDriver)).call(this, '_start', '_stop'));

    _this._running = false;
    _this._port = port;
    _this._teardownHook = false;
    return _this;
  }

  /**
   * @returns {Number}
   */


  _createClass(AbstractDriver, [{
    key: '_registerTeardownHook',


    /**
     * @returns {AbstractDriver}
     * @private
     */
    value: function _registerTeardownHook() {
      if (this._teardownHook) {
        return this;
      }

      this._teardownHook = true;

      process.once('uncaughtException', error => {
        this.stop(() => '');

        throw error;
      });

      process.once('exit', () => {
        this.stop(() => '');
      });

      return this;
    }

    /**
     * @todo: better way to control the timeout of server startup...
     *
     * @param {Number} tts
     * @param {Function} cb
     * @private
     */

  }, {
    key: '_triggerOnTtsExpired',
    value: function _triggerOnTtsExpired(tts, cb) {
      setTimeout(() => {
        if (!this._running) {
          cb();
        }
      }, tts * 1000);
    }

    /**
     * @param {Function} cb
     * @param {Number} tts
     * @returns {AbstractDriver}
     */

  }, {
    key: 'start',
    value: function start(cb) {
      var _this2 = this;

      let tts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : AbstractDriver.DEFAULT_TTS;

      if (this._running) {
        cb(new _ServerAlreadyRunningException.ServerAlreadyRunningException(this));

        return this;
      }

      let _cbTriggered = false;

      this._start(function (error) {
        for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
          args[_key - 1] = arguments[_key];
        }

        _cbTriggered = true;

        if (!error) {
          _this2._running = true;
          _this2._registerTeardownHook();
        }

        cb.apply(undefined, [error].concat(args));
      });

      this._triggerOnTtsExpired(tts, () => {
        if (!_cbTriggered) {
          _cbTriggered = true;

          cb(new _ServerTtsExceededException.ServerTtsExceededException(this, tts));
        }
      });

      return this;
    }

    /**
     * @param {Function} cb
     * @returns {AbstractDriver}
     */

  }, {
    key: 'stop',
    value: function stop(cb) {
      var _this3 = this;

      if (!this._running) {
        cb(null);

        return this;
      }

      this._stop(function (error) {
        for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
          args[_key2 - 1] = arguments[_key2];
        }

        if (!error) {
          _this3._running = false;
        }

        cb.apply(undefined, [error].concat(args));
      });

      return this;
    }

    /**
     * @param {Function} cb
     * @returns {AbstractDriver}
     */

  }, {
    key: 'restart',
    value: function restart(cb) {
      this.stop(error => {
        if (error) {
          cb(error);

          return;
        }

        this.start(cb);
      });

      return this;
    }

    /**
     * @todo: decrease it when local-dynamo fixed
     *
     * @returns {Number}
     */

  }, {
    key: 'port',
    get: function get() {
      return this._port;
    }

    /**
     * @param {Number} port
     */
    ,
    set: function set(port) {
      this._port = port;
    }

    /**
     * @returns {Boolean}
     */

  }, {
    key: 'running',
    get: function get() {
      return this._running;
    }
  }], [{
    key: 'DEFAULT_TTS',
    get: function get() {
      return 60;
    }

    /**
     * @returns {Number}
     */

  }, {
    key: 'DEFAULT_PORT',
    get: function get() {
      return 8877;
    }
  }]);

  return AbstractDriver;
}(_deepCore2.default.OOP.Interface);