/**
 * Created by AlexanderC on 9/23/15.
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.LocalDynamo = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _PathAwareDriver2 = require('./PathAwareDriver');

var _localDynamo = require('local-dynamo');

var _localDynamo2 = _interopRequireDefault(_localDynamo);

var _FailedToStartServerException = require('./Exception/FailedToStartServerException');

var _fsExtra = require('fs-extra');

var _fsExtra2 = _interopRequireDefault(_fsExtra);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

let LocalDynamo = exports.LocalDynamo = function (_PathAwareDriver) {
  _inherits(LocalDynamo, _PathAwareDriver);

  /**
   * @param {Object} options
   * @param {String} path
   * @param {Number} port
   */
  function LocalDynamo() {
    let options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : LocalDynamo.DEFAULT_OPTIONS;
    let path = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : LocalDynamo.DBPath;
    let port = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : LocalDynamo.DEFAULT_PORT;

    _classCallCheck(this, LocalDynamo);

    var _this = _possibleConstructorReturn(this, (LocalDynamo.__proto__ || Object.getPrototypeOf(LocalDynamo)).call(this, path, port));

    _this._options = options;
    _this._process = null;

    _this._pickUpOldInstance = true; // @todo: set it false by default?
    return _this;
  }

  /**
   * @returns {Boolean}
   */


  _createClass(LocalDynamo, [{
    key: '_start',


    /**
     * @param {Function} cb
     * @private
     */
    value: function _start(cb) {
      let cbTriggered = false;

      // avoid local-dynamo package issues...
      _fsExtra2.default.ensureDirSync(this.path);

      this._options.dir = this.path;

      try {
        this._process = _localDynamo2.default.launch(this._options, this.port);
      } catch (error) {
        this._pickUpOldInstance ? cb(null) : cb(new _FailedToStartServerException.FailedToStartServerException(this, error));

        return;
      }

      // This hook fixes DynamoDB startup delay by waiting an empty stdout dataset
      // @todo: remove this hook after fixing issue!
      this._process.stdout.on('data', data => {
        if (data.toString() && !cbTriggered) {
          cbTriggered = true;
          cb(null);
        }
      });

      let onError = error => {
        this._stop(() => {});

        if (!cbTriggered) {
          cbTriggered = true;
          cb(new _FailedToStartServerException.FailedToStartServerException(this, error));
        }
      };

      this._process.on('uncaughtException', onError);
      this._process.on('error', onError);
      this._process.on('exit', () => {
        this.stop(() => '');
      });
    }

    /**
     * @param {Function} cb
     * @private
     */

  }, {
    key: '_stop',
    value: function _stop(cb) {
      if (this._process) {
        this._process.kill();
        this._process = null;
      }

      cb(null);
    }

    /**
     * @returns {Object}
     */

  }, {
    key: 'pickUpOldInstance',
    get: function get() {
      return this._pickUpOldInstance;
    }

    /**
     * @param {Boolean} state
     */
    ,
    set: function set(state) {
      this._pickUpOldInstance = state;
    }

    /**
     * @returns {Object}
     */

  }, {
    key: 'options',
    get: function get() {
      return this._options;
    }
  }], [{
    key: 'DEFAULT_OPTIONS',
    get: function get() {
      return {
        stdio: 'pipe'
      };
    }
  }]);

  return LocalDynamo;
}(_PathAwareDriver2.PathAwareDriver);