/**
 * Created by Stefan Hariton on 6/26/15.
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ExtendModel = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _UndefinedMethodException = require('./Exceptions/UndefinedMethodException');

var _crypto = require('crypto');

var _crypto2 = _interopRequireDefault(_crypto);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Extends standard Vogels models
 */
let ExtendModel = exports.ExtendModel = function () {
  /**
   * @param {Object} model
   */
  function ExtendModel(model) {
    _classCallCheck(this, ExtendModel);

    this._model = model;

    this._registerQueryWrappers();
  }

  /**
   * @returns {ExtendModel}
   * @private
   */


  _createClass(ExtendModel, [{
    key: '_registerQueryWrappers',
    value: function _registerQueryWrappers() {
      this._model.setPartition = function (partition) {
        this[ExtendModel.PARTITION_KEY] = partition;
      };

      this._model.getPartition = function () {
        if (this.hasOwnProperty(ExtendModel.ONE_TIME_PARTITION_KEY)) {
          let partition = this[ExtendModel.ONE_TIME_PARTITION_KEY];

          delete this[ExtendModel.ONE_TIME_PARTITION_KEY];

          return partition;
        } else if (this.hasOwnProperty(ExtendModel.PARTITION_KEY)) {
          return this[ExtendModel.PARTITION_KEY];
        }

        return null;
      };

      this._model.hasPartition = function () {
        return this.hasOwnProperty(ExtendModel.ONE_TIME_PARTITION_KEY) || this.hasOwnProperty(ExtendModel.PARTITION_KEY);
      };

      this._model.useAnonymous = function () {
        this[ExtendModel.ONE_TIME_PARTITION_KEY] = ExtendModel.ANONYMOUS_PARTITION;

        return this;
      };

      this._model.deepQuery = function () {
        let requestStrategy = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : ExtendModel.QUERY_STRATEGY;

        if (this.hasPartition()) {
          switch (requestStrategy) {
            case ExtendModel.SCAN_STRATEGY:
              return this.scan().where(ExtendModel.PARTITION_FIELD).equals(this.getPartition());
            case ExtendModel.QUERY_STRATEGY:
              return this.query(this.getPartition());
          }
        } else {
          return this.scan();
        }
      };

      this._model.anonymousQuery = function () {
        return this.query(ExtendModel.ANONYMOUS_PARTITION);
      };

      return this;
    }

    /**
     * @returns {Object}
     */

  }, {
    key: 'inject',


    /**
     * Injects the specified methods or all
     * @param {Array} methods
     * @returns {Object|*}
     */
    value: function inject() {
      let methods = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

      let predefinedMethods = this.methods;
      let predefinedMethodsNames = Object.keys(predefinedMethods);

      methods = methods || predefinedMethodsNames;

      for (let methodKey in methods) {
        if (!methods.hasOwnProperty(methodKey)) {
          continue;
        }

        let methodName = methods[methodKey];

        if (!predefinedMethods.hasOwnProperty(methodName)) {
          throw new _UndefinedMethodException.UndefinedMethodException(methodName, predefinedMethodsNames);
        }

        this._model[methodName] = this._applyRumDecorator(predefinedMethods[methodName], methodName);
      }

      return this._model;
    }

    /**
     * @param {Object} customData
     * @returns {Boolean}
     * @private
     */

  }, {
    key: '_logRumEvent',
    value: function _logRumEvent(customData) {
      if (!this.model.logService) {
        return false;
      }

      let event = Object.assign(customData, {
        service: 'deep-db',
        resourceType: 'DynamoDB',
        resourceId: this.model.tableName()
      });

      this.model.logService.rumLog(event);

      return true;
    }

    /**
     * @param {Function} method
     * @param {String} eventName
     * @returns {Function}
     * @private
     */

  }, {
    key: '_applyRumDecorator',
    value: function _applyRumDecorator(method, eventName) {
      var _this2 = this;

      return function () {
        for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
          args[_key] = arguments[_key];
        }

        if (!_this2.model.logService) {
          return method.apply(undefined, args);
        }

        let cbIndex = null;

        for (let index in args) {
          if (args.hasOwnProperty(index) && typeof args[index] === 'function') {
            cbIndex = index;
          }
        }

        if (cbIndex === null) {
          return method.apply(undefined, args);
        }

        let startOpEvent = {
          eventName: eventName,
          time: Date.now(),
          eventId: ExtendModel.buildEventId(eventName)
        };

        _this2._logRumEvent(startOpEvent);

        let originalCb = args[cbIndex];

        args[cbIndex] = function () {
          let endOpEvent = Object.assign({}, startOpEvent);
          endOpEvent.time = Date.now();

          _this2._logRumEvent(endOpEvent);

          originalCb.apply(undefined, arguments);
        };

        return method.apply(undefined, args);
      };
    }

    /**
     * @param {String} eventName
     * @returns {String}
     */

  }, {
    key: 'model',
    get: function get() {
      return this._model;
    }

    /**
     * @returns {Number}
     */

  }, {
    key: 'methods',


    /**
     * @returns {Object}
     */
    get: function get() {
      let _this = this;

      return {

        // Avoid 3'rd party errors (unhandled rejections) 
        // by wrapping the callback in a setImmediate
        _findUntilLimitCb: function _findUntilLimitCb(cb) {
          var _this$model;

          for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
            args[_key2 - 1] = arguments[_key2];
          }

          (_this$model = _this.model)._findUntilLimit.apply(_this$model, _toConsumableArray(args)).then(result => setImmediate(() => cb(null, result))).catch(error => setImmediate(() => cb(error, null)));
        },


        // Fixes DynamoDB limit behavior
        // @ref http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/QueryAndScan.html#ScanQueryLimit
        _findUntilLimit: function _findUntilLimit(query, limit) {
          let _scannedCount = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;

          let _accumulator = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : [];

          let _lastKey = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : null;

          return new Promise((resolve, reject) => {
            if (_lastKey) {
              query.startKey(_lastKey);
            }

            query.limit(limit * 100) // DDB native limit (#ExclusiveStartKey)
            .exec((error, result) => {
              if (error) {
                return reject(error);
              }

              _accumulator = _accumulator.concat(result.Items || []);
              _lastKey = result.LastEvaluatedKey || null;
              _scannedCount += result.ScannedCount;

              if (!result || !_lastKey || result.ScannedCount <= 0 || _accumulator.length >= limit) {
                while (_accumulator.length > limit) {
                  _accumulator.pop();
                }

                const finalResult = {
                  ScannedCount: _scannedCount,
                  Count: _accumulator.length,
                  Items: _accumulator
                };

                if (result.ConsumedCapacity) {
                  finalResult.ConsumedCapacity = result.ConsumedCapacity;
                }

                return resolve(finalResult);
              }

              _this.model._findUntilLimit(query, limit, _scannedCount, _accumulator, _lastKey).then(resolve).catch(reject);
            });
          });
        },


        findAll: function findAll(cb) {
          return _this.model.deepQuery().loadAll().exec(cb);
        },

        findAllPaginated: function findAllPaginated(startKey, limit, cb) {
          return _this.model._findUntilLimitCb(cb, _this.model.deepQuery(), limit, 0, [], startKey);
        },

        findOneById: function findOneById(id, cb) {
          return _this.model.hasPartition() ? _this.model.get(_this.model.getPartition(), id, cb) : _this.model.get(id, cb);
        },

        findOneBy: function findOneBy(fieldName, value, cb) {
          return _this.model.findBy(fieldName, value, cb, 1);
        },

        findBy: function findBy(fieldName, value, cb) {
          let limit = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : ExtendModel.DEFAULT_LIMIT;

          const query = _this.model.deepQuery(ExtendModel.SCAN_STRATEGY).where(fieldName).equals(value);

          return _this.model._findUntilLimitCb(cb, query, limit);
        },

        findAllBy: function findAllBy(fieldName, value, cb) {
          return _this.model.deepQuery(ExtendModel.SCAN_STRATEGY).where(fieldName).equals(value).loadAll().exec(cb);
        },

        findAllByPaginated: function findAllByPaginated(fieldName, value, startKey, limit, cb) {
          return _this.model.deepQuery(ExtendModel.SCAN_STRATEGY).where(fieldName).equals(value).startKey(startKey).limit(limit).exec(cb);
        },

        findOneMatching: function findOneMatching(params, cb) {
          return _this.model.findMatching(params, cb, 1);
        },

        findMatching: function findMatching(params, cb) {
          let limit = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : ExtendModel.DEFAULT_LIMIT;

          let scanParams = ExtendModel.buildScanParameters(params);

          const query = _this.model.deepQuery().filterExpression(scanParams.filterExpression).expressionAttributeValues(scanParams.filterExpressionValues).expressionAttributeNames(scanParams.filterExpressionNames);

          return _this.model._findUntilLimitCb(cb, query, limit);
        },

        findAllMatching: function findAllMatching(params, cb) {
          let scanParams = ExtendModel.buildScanParameters(params);

          return _this.model.deepQuery().filterExpression(scanParams.filterExpression).expressionAttributeValues(scanParams.filterExpressionValues).expressionAttributeNames(scanParams.filterExpressionNames).loadAll().exec(cb);
        },

        findAllMatchingPaginated: function findAllMatchingPaginated(params, startKey, limit, cb) {
          let scanParams = ExtendModel.buildScanParameters(params);

          return _this.model.deepQuery().filterExpression(scanParams.filterExpression).expressionAttributeValues(scanParams.filterExpressionValues).expressionAttributeNames(scanParams.filterExpressionNames).startKey(startKey).limit(limit).exec(cb);
        },

        findItems: function findItems() {
          var _this$model2;

          for (var _len3 = arguments.length, args = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
            args[_key3] = arguments[_key3];
          }

          if (_this.model.hasPartition()) {
            let partition = _this.model.getPartition();
            let ids = args.shift();
            let composedIds = ids.map(id => {
              let idObj = {};

              idObj[ExtendModel.PARTITION_FIELD] = partition;
              idObj.Id = id;

              return idObj;
            });

            // concat with anonymous partition search
            if (partition !== ExtendModel.ANONYMOUS_PARTITION) {
              composedIds = composedIds.concat(ids.map(id => {
                let idObj = {};

                idObj[ExtendModel.PARTITION_FIELD] = ExtendModel.ANONYMOUS_PARTITION;
                idObj.Id = id;

                return idObj;
              }));
            }

            args.unshift(composedIds);
          }

          return (_this$model2 = _this.model).getItems.apply(_this$model2, args);
        },

        deleteById: function deleteById(id, cb) {
          return _this.model.hasPartition() ? _this.model.destroy(_this.model.getPartition(), id, cb) : _this.model.destroy(id, cb);
        },

        deleteByIdConditional: function deleteByIdConditional(id, condition, cb) {
          return _this.model.destroy(id, condition, cb);
        },

        createItem: function createItem(data, cb) {
          if (_this.model.hasPartition()) {
            data[ExtendModel.PARTITION_FIELD] = _this.model.getPartition();
          }

          return _this.model.create(data, cb);
        },

        createUniqueOnFields: function createUniqueOnFields(fields, data, cb) {
          let scanCb = function scanCb(err, data) {
            if (err) {
              return cb(err, data);
            }

            if (data.Count) {
              return cb(`Item like ${data} already exists`);
            }

            return _this.model.create(data, cb);
          };

          let scanParams = {};
          for (let fieldKey in fields) {
            if (!fields.hasOwnProperty(fieldKey)) {
              continue;
            }

            let field = fields[fieldKey];

            scanParams[field] = data[field];
          }

          scanParams = ExtendModel.buildScanParameters(scanParams);

          return _this.model.deepQuery().filterExpression(scanParams.filterExpression).expressionAttributeValues(scanParams.filterExpressionValues).expressionAttributeNames(scanParams.filterExpressionNames).limit(1).exec(scanCb);
        },

        updateItem: function updateItem(id, data, cb) {
          data.Id = id;

          if (_this.model.hasPartition()) {
            data[ExtendModel.PARTITION_FIELD] = _this.model.getPartition();
          }

          return _this.model.update(data, cb);
        },

        updateItemConditional: function updateItemConditional(id, data, condition, cb) {
          data.Id = id;

          if (_this.model.hasPartition()) {
            data[ExtendModel.PARTITION_FIELD] = _this.model.getPartition();
          }

          return _this.model.update(data, condition, cb);
        }
      };
    }
  }], [{
    key: 'buildScanParameters',


    /**
     * Makes filterExpression, filtersExpressionValues and filterExpressionNames from an object, that are used to make
     * a DynamoDb scan
     *
     * @param {Object} params
     * @returns {Object}
     */
    value: function buildScanParameters(params) {
      let filterExpression = '';
      let filterExpressionValues = {};
      let filterExpressionNames = {};
      let first = true;

      for (let key in params) {
        if (!params.hasOwnProperty(key)) {
          continue;
        }

        let fieldValue = params[key];

        let fieldName = `#${key}`;
        let fieldValueName = `:${key}`;

        if (!first) {
          filterExpression += ' AND ';
        }

        filterExpression += `${fieldName} = ${fieldValueName}`;
        filterExpressionValues[fieldValueName] = fieldValue;
        filterExpressionNames[fieldName] = key;
        first = false;
      }

      return {
        filterExpression: filterExpression,
        filterExpressionValues: filterExpressionValues,
        filterExpressionNames: filterExpressionNames
      };
    }
  }, {
    key: 'buildEventId',
    value: function buildEventId(eventName) {
      var md5sum = _crypto2.default.createHash('md5');

      md5sum.update(JSON.stringify({
        namespace: 'DB|Vogels|ExtendModel',
        name: eventName,
        time: Date.now()
      }));

      return md5sum.digest('hex');
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'DEFAULT_LIMIT',
    get: function get() {
      return 10;
    }

    /**
     * @returns {number}
     */

  }, {
    key: 'DEFAULT_SEGMENTS_NUMBER',
    get: function get() {
      return 4;
    }
  }, {
    key: 'PARTITION_FIELD',
    get: function get() {
      return 'AccountId';
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'ONE_TIME_PARTITION_KEY',
    get: function get() {
      return 'oneTimeDeepPartitionKey';
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'PARTITION_KEY',
    get: function get() {
      return 'deepPartitionKey';
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'ANONYMOUS_PARTITION',
    get: function get() {
      return 'anonymous';
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'SCAN_STRATEGY',
    get: function get() {
      return 'scan';
    }

    /**
     * @returns {String}
     */

  }, {
    key: 'QUERY_STRATEGY',
    get: function get() {
      return 'query';
    }
  }]);

  return ExtendModel;
}();