/**
 * Created by AlexanderC on 9/23/15.
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.DBServer = undefined;

var _LocalDynamo = require('./Driver/LocalDynamo');

var _Dynalite = require('./Driver/Dynalite');

class DBServer {
  /**
   * @param {Function|String} driver
   * @param {*} args
   * @returns {Function}
   */
  static create(driver = DBServer.DEFAULT_DRIVER, ...args) {
    let DriverProto = typeof driver === 'string' ? DBServer._findDriverPrototype(driver) : driver;

    if (!DriverProto) {
      throw new Error(`Missing DB server driver ${driver}`);
    }

    return new DriverProto(...args);
  }

  /**
   * @param {String} name
   * @returns {Function}
   * @private
   */
  static _findDriverPrototype(name) {
    for (let driverProtoKey in DBServer.DRIVERS) {
      if (!DBServer.DRIVERS.hasOwnProperty(driverProtoKey)) {
        continue;
      }

      let driverProto = DBServer.DRIVERS[driverProtoKey];

      if (driverProto.name === name) {
        return driverProto;
      }
    }

    return null;
  }

  /**
   * @returns {Function[]}
   */
  static get DRIVERS() {
    return [_LocalDynamo.LocalDynamo, _Dynalite.Dynalite];
  }

  /**
   * @returns {Function}
   */
  static get DEFAULT_DRIVER() {
    return DBServer.DRIVERS[0];
  }
}
exports.DBServer = DBServer;