/**
 * Created by AlexanderC on 9/23/15.
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Dynalite = undefined;

var _PathAwareDriver = require('./PathAwareDriver');

var _dynalite = require('dynalite');

var _dynalite2 = _interopRequireDefault(_dynalite);

var _FailedToStartServerException = require('./Exception/FailedToStartServerException');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Dynalite extends _PathAwareDriver.PathAwareDriver {
  /**
   * @param {Object} options
   * @param {String} path
   * @param {Number} port
   */
  constructor(options = Dynalite.DEFAULT_OPTIONS, path = Dynalite.DBPath, port = Dynalite.DEFAULT_PORT) {
    super(path, port);

    this._options = options;
    this._server = null;
  }

  /**
   * @returns {Object}
   */
  get options() {
    return this._options;
  }

  /**
   * @param {Function} cb
   * @private
   */
  _start(cb) {
    this._options.path = this.path;

    this._server = (0, _dynalite2.default)(this._options);

    this._server.listen(this.port, error => {
      if (error) {
        cb(new _FailedToStartServerException.FailedToStartServerException(this, error));
        return;
      }

      cb(null);
    });
  }

  /**
   * @param {Function} cb
   * @private
   */
  _stop(cb) {
    if (this._server) {
      this._server.close(cb);
      this._server = null;

      return;
    }

    cb(null);
  }

  /**
   * @returns {Object}
   */
  static get DEFAULT_OPTIONS() {
    return {
      createTableMs: 0,
      deleteTableMs: 0,
      updateTableMs: 0
    };
  }
}
exports.Dynalite = Dynalite;